#encoding=utf8

__version__ = '0.3.4'
__all__ = [
    'MLBase',
    'PLA',
    'PLA.BinaryClassifier',
    'PLA.MultiClassifier',
    'PocketPLA',
    'PocketPLA.BinaryClassifier',
    'PocketPLA.MultiClassifier',
    'LinearRegression',
    'LinearRegression.LinearRegression',
    'LinearRegression.BinaryClassifier',
    'LinearRegression.MultiClassifier',
    'LinearRegression.Accelerator',
    'RidgeRegression',
    'RidgeRegression.RidgeRegression',
    'RidgeRegression.BinaryClassifier',
    'RidgeRegression.MultiClassifier',
    'KernelRidgeRegression',
    'KernelRidgeRegression.KernelRidgeRegression',
    'KernelRidgeRegression.BinaryClassifier',
    'LogisticRegression',
    'LogisticRegression.LogisticRegression',
    'LogisticRegression.BinaryClassifier',
    'LogisticRegression.MultiClassifier',
    'L2RLogisticRegression',
    'L2RLogisticRegression.L2RLogisticRegression',
    'L2RLogisticRegression.BinaryClassifier',
    'KernelLogisticRegression',
    'KernelLogisticRegression.KernelLogisticRegression',
    'SupportVectorMachine',
    'SupportVectorMachine.BinaryClassifier',
    'SupportVectorMachine.MultiClassifier',
    'ProbabilisticSVM',
    'ProbabilisticSVM.ProbabilisticSVM',
    'LeastSquaresSVM',
    'LeastSquaresSVM.BinaryClassifier',
    'SupportVectorRegression',
    'SupportVectorRegression.SupportVectorRegression',
    'DecisionStump',
    'DecisionStump.BinaryClassifier',
    'AdaBoostStump',
    'AdaBoostStump.BinaryClassifier',
    'DecisionTree',
    'DecisionTree.DecisionTree',
    'DecisionTree.CART',
    'Blending',
    'Blending.UniformBlendingClassifier',
    'Blending.LinearBlendingClassifier',
    'Utility',
    'Utility.DatasetLoader',
    'Utility.CrossValidator',
    'Utility.Kernel',
    'Utility.Serializer'
]
